#include <assert.h>
#include <string.h>
#include "x_err.hpp"
#ifdef N_COMPRESSORS
#include "x_huff.hpp"
#endif
#include "unique.hpp"
#include "packet.hpp"
#include "compress.hpp"

#ifdef N_COMPRESSORS
Huff			*H[N_COMPRESSORS];
#endif
int                     cPacketLen;
byte			cPacket[MAX_MODE_LEN * 9 / 8 + 1];

static byte	        last2Sent[MAX_NETNODES][MAX_MODE_LEN][2];
static byte		last2Rcvd[MAX_NETNODES][MAX_MODE_LEN][2];
#ifdef DEBUG
static dword		nCompressed[2], nUncompressed;
static dword		compressedLen[2][MAX_MODE_LEN * 9 / 8 + 1];
static dword		bits[MAX_MODE_LEN], bytes[MAX_MODE_LEN];
static word             maxCompressedLen;
#endif

#ifdef DEBUG
void			compressDebug();
#endif
void			compressPacket();
int			uncompressPacket(byte *buf);
void                    compressInit(int node, int type);

#ifdef DEBUG
void compressDebug() {
  dword i, j;
  FILE *f = fopen("debug.txt", "at");
  fprintf(f, "--- COMPRESS Module Debug Info ---------------------------------\n");
  fprintf(f, "nCompressed[0]:           %lu\n", nCompressed[0]);
  fprintf(f, "nCompressed[1]:           %lu\n", nCompressed[1]);
  fprintf(f, "nUncompressed:            %lu\n", nUncompressed);
  fprintf(f, "maxCompressedLen:         %u\n", maxCompressedLen);
  if (nCompressed[0]) {
    for (j = 0, i = 0; i < MAX_MODE_LEN * 9 / 8 + 2; i++)
      j += i * compressedLen[0][i];
    fprintf(f, "mean compressed length:   %.2f\n", (float)j / nCompressed[0]);
  }
  if (nCompressed[1]) {
    for (j = 0, i = 0; i < MAX_MODE_LEN * 9 / 8 + 2; i++)
      j += i * compressedLen[1][i];
    fprintf(f, "mean compressed length 2: %.2f\n", (float)j / nCompressed[1]);
  }
  for (i = 0; i < MAX_MODE_LEN; i++)
    if (bytes[i])
      fprintf(f, "%lu: %lu / %lu (%.2f%%)\n", i, bits[i], bytes[i] * 8, (float)bits[i] / (bytes[i] * 8) * 100);
  fclose(f);
}
#endif

void compressPacket() {
#ifdef N_COMPRESSORS
  int len;
  ushort code;
#endif
  int i, j, *cMethod;
#ifdef RLE_BITS
  int runLength, RLEend, runEnds;
#endif
  ushort iMask, oMask;
  oMask = 0x80; cPacketLen = 0;
#ifdef COMPRESS_RANGE
  cMethod = (int*)::cMethod;
#if COMPRESS_RANGE_BITS > 8
  cPacket[0] = (packetLen - 1) & 0xff;
  cPacket[1] = ((packetLen - 1) & 0xff00) >> (COMPRESS_RANGE_BITS - 8);
  oMask >>= COMPRESS_RANGE_BITS - 8;
  cPacketLen++;
#else
  cPacket[0] = (packetLen - 1) << (8 - COMPRESS_RANGE_BITS);
#if COMPRESS_RANGE_BITS == 8
  cPacket[++cPacketLen] = 0;
#else
  oMask >>= COMPRESS_RANGE_BITS;
#endif
#endif
#else
  if (packetLen == MODE_LEN) {
    cMethod = (int*)::cMethod;
    cPacket[0] = 0;
  } else {
    cMethod = (int*)::cMethod + MODE_LEN;
    cPacket[0] = oMask;
  }
  oMask >>= 1;
#endif
  for (i = 0; i < packetLen; i++) {
    switch (cMethod[i]) {
      case CM_NONE:
#ifdef DEBUG
        bytes[i]++;
#endif
	for (iMask = 0x80; iMask; iMask >>= 1) {
	  if (packet[i] & iMask) cPacket[cPacketLen] |= oMask;
	  oMask >>= 1;
	  if (!oMask) {
	    oMask = 0x80;
	    cPacket[++cPacketLen] = 0;
	  }
	}
#ifdef DEBUG
	bits[i] += 8;
#endif
	break;
      case CM_RUN:
#ifdef DEBUG
        bytes[i]++;
#endif
	if (last2Sent[remoteAddr][i][0] == last2Sent[remoteAddr][i][1]) {
#ifdef DEBUG
	  bits[i]++;
#endif
	  if (packet[i] == last2Sent[remoteAddr][i][0]) {
	    cPacket[cPacketLen] |= oMask;
	    oMask >>= 1;
	    if (!oMask) {
	      oMask = 0x80;
	      cPacket[++cPacketLen] = 0;
	    }
	    break;
	  } else {
	    oMask >>= 1;
	    if (!oMask) {
	      oMask = 0x80;
	      cPacket[++cPacketLen] = 0;
	    }
	  }
	}
	for (iMask = 0x80; iMask; iMask >>= 1) {
	  if (packet[i] & iMask) cPacket[cPacketLen] |= oMask;
	  oMask >>= 1;
	  if (!oMask) {
	    oMask = 0x80;
	    cPacket[++cPacketLen] = 0;
	  }
	}
#ifdef DEBUG
	bits[i] += 8;
#endif
	break;
      case CM_INC:
#ifdef DEBUG
        bytes[i]++;
#endif
        if (last2Sent[remoteAddr][i][1] == last2Sent[remoteAddr][i][0] + 1) {
#ifdef DEBUG
	  bits[i]++;
#endif
	  if (packet[i] == last2Sent[remoteAddr][i][1] + 1) {
	    cPacket[cPacketLen] |= oMask;
	    oMask >>= 1;
	    if (!oMask) {
	      oMask = 0x80;
	      cPacket[++cPacketLen] = 0;
	    }
	    break;
	  } else {
	    oMask >>= 1;
	    if (!oMask) {
	      oMask = 0x80;
	      cPacket[++cPacketLen] = 0;
	    }
	  }
	}
	for (iMask = 0x80; iMask; iMask >>= 1) {
	  if (packet[i] & iMask) cPacket[cPacketLen] |= oMask;
	  oMask >>= 1;
	  if (!oMask) {
	    oMask = 0x80;
	    cPacket[++cPacketLen] = 0;
	  }
	}
#ifdef DEBUG
	bits[i] += 8;
#endif
	break;
      case CM_DEC:
#ifdef DEBUG
        bytes[i]++;
#endif
        if (last2Sent[remoteAddr][i][1] == last2Sent[remoteAddr][i][0] - 1) {
#ifdef DEBUG
	  bits[i]++;
#endif
	  if (packet[i] == last2Sent[remoteAddr][i][1] - 1) {
	    cPacket[cPacketLen] |= oMask;
	    oMask >>= 1;
	    if (!oMask) {
	      oMask = 0x80;
	      cPacket[++cPacketLen] = 0;
	    }
	    break;
	  } else {
	    oMask >>= 1;
	    if (!oMask) {
	      oMask = 0x80;
	      cPacket[++cPacketLen] = 0;
	    }
	  }
	}
	for (iMask = 0x80; iMask; iMask >>= 1) {
	  if (packet[i] & iMask) cPacket[cPacketLen] |= oMask;
	  oMask >>= 1;
	  if (!oMask) {
	    oMask = 0x80;
	    cPacket[++cPacketLen] = 0;
	  }
	}
#ifdef DEBUG
	bits[i] += 8;
#endif
	break;
#ifdef RLE_BITS
      case CM_RLE_S:
	j = i;
	runLength = RLEend = 0;
	while (!RLEend) {
#ifdef DEBUG
          bytes[j]++;
#endif
	  switch (cMethod[j]) {
	    case CM_RLE_S:
	      goto l1;
	    case CM_RLE_E:
	      RLEend = 1;
	    case CM_RLE:
l1:           if (j == packetLen - 1) RLEend = 1;
              if (last2Sent[remoteAddr][j][0] == last2Sent[remoteAddr][j][1] &&
                  last2Sent[remoteAddr][j][1] == packet[j])
              {
		runLength++;
		runEnds = 0;
	      } else {
		runEnds = 1;
	      }
	      if (runLength && (runEnds || RLEend || runLength == RLE_LEN)) {
#ifdef DEBUG
		bits[j] += 1 + RLE_BITS;
#endif
		oMask >>= 1;
		if (!oMask) {
		  oMask = 0x80;
		  cPacket[++cPacketLen] = 0;
		}
		runLength--;
		for (iMask = 1 << (RLE_BITS - 1); iMask; iMask >>= 1) {
		  if (runLength & iMask) cPacket[cPacketLen] |= oMask;
		  oMask >>= 1;
		  if (!oMask) {
		    oMask = 0x80;
		    cPacket[++cPacketLen] = 0;
		  }
		}
		runLength = 0;
	      }
              if (runEnds) {
#ifdef DEBUG
		bits[j] += 9;
#endif
		cPacket[cPacketLen] |= oMask;
		oMask >>= 1;
		if (!oMask) {
		  oMask = 0x80;
		  cPacket[++cPacketLen] = 0;
		}
		for (iMask = 0x80; iMask; iMask >>= 1) {
		  if (packet[j] & iMask) cPacket[cPacketLen] |= oMask;
		  oMask >>= 1;
		  if (!oMask) {
		    oMask = 0x80;
		    cPacket[++cPacketLen] = 0;
		  }
		}
	      }
            default:
              if (j == packetLen - 1 && runLength) {
#ifdef DEBUG
		bits[j] += 1 + RLE_BITS;
#endif
		oMask >>= 1;
		if (!oMask) {
		  oMask = 0x80;
		  cPacket[++cPacketLen] = 0;
		}
		runLength--;
		for (iMask = 1 << (RLE_BITS - 1); iMask; iMask >>= 1) {
		  if (runLength & iMask) cPacket[cPacketLen] |= oMask;
		  oMask >>= 1;
		  if (!oMask) {
		    oMask = 0x80;
		    cPacket[++cPacketLen] = 0;
		  }
		}
              }
	  }
	  j++;
	}
	break;
      case CM_RLE:
      case CM_RLE_E:
        break;
#endif
#ifdef N_COMPRESSORS
      default:			// Huffman
#ifdef DEBUG
        bytes[i]++;
#endif
	H[cMethod[i]]->Encode_Symbol(packet[i], &code, &len);
#ifdef DEBUG
	bits[i] += len;
#endif
	for (iMask = 1 << (len - 1); iMask; iMask >>= 1) {
	  if (code & iMask) cPacket[cPacketLen] |= oMask;
	  oMask >>= 1;
	  if (!oMask) {
	    oMask = 0x80;
	    cPacket[++cPacketLen] = 0;
	  }
	}
#endif
    }
    last2Sent[remoteAddr][i][0] = last2Sent[remoteAddr][i][1];
    last2Sent[remoteAddr][i][1] = packet[i];
  }
  if (remoteAddr == nodeAddr) {
    for (i = 0; i <= maxAddr; i++)
      if (nodeExists[i]) for (j = 0; j < packetLen; j++) {
        last2Sent[i][j][0] = last2Sent[i][j][1];
        last2Sent[i][j][1] = last2Sent[nodeAddr][j][1];
      }
  }
  if (oMask != 0x80) cPacketLen++;
#ifdef DEBUG
  nCompressed[packetLen == MODE_LEN2]++;
  compressedLen[packetLen == MODE_LEN2][cPacketLen]++;
  if (cPacketLen > maxCompressedLen) maxCompressedLen = cPacketLen;
#endif
}

int uncompressPacket(byte *buf) {
  byte *data;
  byte mask;
  byte oMask;
#ifdef RLE_BITS
  int k, runLength, RLEend;
#endif
  int i, j, pktLen, *cMethod;
  data = buf;
  mask = 0x80;
#ifdef COMPRESS_RANGE
  cMethod = (int*)::cMethod;
#if COMPRESS_RANGE_BITS > 8
  pktLen = *data;
  data++;
  pktLen |= ((word)*data & ~((1 << (16 - COMPRESS_RANGE_BITS)) - 1))
            << (COMPRESS_RANGE_BITS - 8);
  pktLen++;
  mask >>= COMPRESS_RANGE_BITS - 8;
#else
  pktLen = (*data >> (8 - COMPRESS_RANGE_BITS)) + 1;
#if COMPRESS_RANGE_BITS == 8
  data++;
#else
  mask >>= COMPRESS_RANGE_BITS;
#endif
#endif
  if (pktLen > MAX_MODE_LEN) return 1;
#else
  if (!(*data & mask)) {
    pktLen = MODE_LEN;
    cMethod = (int*)::cMethod;
  } else {
    pktLen = MODE_LEN2;
    cMethod = (int*)::cMethod + MODE_LEN;
  }
  mask >>= 1;
#endif
  for (i = 0; i < pktLen; i++) {
    switch (cMethod[i]) {
      case CM_NONE:
	packet[i] = 0;
	for (oMask = 0x80; oMask; oMask >>= 1) {
	  if (*data & mask) packet[i] |= oMask;
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	}
	break;
      case CM_RUN:
	if (last2Rcvd[remoteAddr][i][0] == last2Rcvd[remoteAddr][i][1]) {
	  j = (*data & mask);
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	  if (j) {
	    packet[i] = last2Rcvd[remoteAddr][i][0];
	    break;
	  }
	}
	packet[i] = 0;
	for (oMask = 0x80; oMask; oMask >>= 1) {
	  if (*data & mask) packet[i] |= oMask;
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	}
	break;
      case CM_INC:
	if (last2Rcvd[remoteAddr][i][1] == last2Rcvd[remoteAddr][i][0] + 1) {
	  j = (*data & mask);
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	  if (j) {
	    packet[i] = last2Rcvd[remoteAddr][i][1] + 1;
	    break;
	  }
	}
	packet[i] = 0;
	for (oMask = 0x80; oMask; oMask >>= 1) {
	  if (*data & mask) packet[i] |= oMask;
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	}
	break;
      case CM_DEC:
	if (last2Rcvd[remoteAddr][i][1] == last2Rcvd[remoteAddr][i][0] - 1) {
	  j = (*data & mask);
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	  if (j) {
	    packet[i] = last2Rcvd[remoteAddr][i][1] - 1;
	    break;
	  }
	}
	packet[i] = 0;
	for (oMask = 0x80; oMask; oMask >>= 1) {
	  if (*data & mask) packet[i] |= oMask;
	  mask >>= 1;
	  if (!mask) {
	    mask = 0x80;
	    data++;
	  }
	}
	break;
#ifdef RLE_BITS
      case CM_RLE_S:
	j = i;
	runLength = RLEend = 0;
	while (!RLEend) {
	  switch (cMethod[j]) {
	    case CM_RLE_S:
	      goto l1;
	    case CM_RLE_E:
	      RLEend = 1;
	    case CM_RLE:
l1:           if (j == pktLen - 1) RLEend = 1;
	      if (!runLength) {
		k = (*data & mask);
		mask >>= 1;
		if (!mask) {
		  mask = 0x80;
		  data++;
		}
		if (k) {
		  packet[j] = 0;
		  for (oMask = 0x80; oMask; oMask >>= 1) {
		    if (*data & mask) packet[j] |= oMask;
		    mask >>= 1;
		    if (!mask) {
		      mask = 0x80;
		      data++;
		    }
		  }
		} else {
		  runLength = 0;
		  for (oMask = 1 << (RLE_BITS - 1); oMask; oMask >>= 1) {
		    if (*data & mask) runLength |= oMask;
		    mask >>= 1;
		    if (!mask) {
		      mask = 0x80;
		      data++;
		    }
		  }
		  runLength++;
		}
	      }
	      if (runLength) {
		packet[j] = last2Rcvd[remoteAddr][j][0];
		runLength--;
		break;
	      }
	  }
	  j++;
	}
	break;
      case CM_RLE:
      case CM_RLE_E:
        break;
#endif
#ifdef N_COMPRESSORS
      default:
	packet[i] = H[cMethod[i]]->Decode_Symbol(&mask, &data);
#endif
    }
    last2Rcvd[remoteAddr][i][0] = last2Rcvd[remoteAddr][i][1];
    last2Rcvd[remoteAddr][i][1] = packet[i];
  }
  packetLen = pktLen;
#ifdef DEBUG
  nUncompressed++;
#endif
  return 0;
}

void compressInit(int node, int type) {
  int i, j;
  if (type == CTRL_RESET_REQUEST) {
    if (node == nodeAddr)
      for (i = 0; i <= maxAddr; i++)
        if (nodeExists[i])
	  for (j = 0; j < MAX_MODE_LEN; j++)
	    last2Sent[i][j][0] = last2Sent[i][j][1] = 0;
    for (i = 0; i < MAX_MODE_LEN; i++)
      last2Sent[node][i][0] = last2Sent[node][i][1] = 0;
  } else {
    for (i = 0; i < MAX_MODE_LEN; i++)
      last2Rcvd[node][i][0] = last2Rcvd[node][i][1] = 0;
  }
}
